/* ***** BEGIN LICENSE BLOCK Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 *
 * The Original Code is TotalQuickFilter.
 *
 * The Initial Developer of the Original Code is alta88 <alta88@gmail.com>.
 
 * Portions created by the Initial Developer are Copyright (C) 2015 the Initial
 * Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file
 * may be used under the terms of either the GNU General Public License Version
 * 2 or later (the "GPL"), or the GNU Lesser General Public License Version 2.1
 * or later (the "LGPL"), in which case the provisions of the GPL or the LGPL
 * are applicable instead of those above. If you wish to allow use of your
 * version of this file only under the terms of either the GPL or the LGPL, and
 * not to allow others to use your version of this file under the terms of the
 * MPL, indicate your decision by deleting the provisions above and replace them
 * with the notice and other provisions required by the GPL or the LGPL. If you
 * do not delete the provisions above, a recipient may use your version of this
 * file under the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */

Components.utils.import("resource:///modules/gloda/log4moz.js");
Components.utils.import("resource://gre/modules/AddonManager.jsm");

var TotalQuickFilter = {
  _id: "totalquickfilter@mozdev.org",
  prefBranch: "extensions.TotalQuickFilter",

  d: function(elementId)   { return document.getElementById(elementId); },

  get log() {
    return Log4Moz.getConfiguredLogger(this.prefBranch);
  },

  searchStrings: Services.strings.createBundle(
      "chrome://messenger/locale/search-attributes.properties"),

  onLoad: function() {
    setTimeout(function() {
TotalQuickFilter.log.debug("onLoad:  ");
      TotalQuickFilter.init();
      window.addEventListener("beforecustomization",
                              TotalQuickFilter.beforeCustomization);
      window.addEventListener("aftercustomization",
                              TotalQuickFilter.afterCustomization);
    }, 0);
  },

  init: function() {
TotalQuickFilter.log.debug("init: START ");
//    TotalQuickFilter.initDateButtons();
    TotalQuickFilter.initTextButtons();
    TotalQuickFilter.initButtonsFunc("tqfButtonIcon");
    TotalQuickFilter.setTagBooleanButton(null);
//    let b = TotalQuickFilter.initTextFilterButtonFunc("tqf-date-filter");
//    TotalQuickFilter.setTextFilterBarButton(b, "tqf-date-filter");
    let b = TotalQuickFilter.initTextFilterButtonFunc("qfb-text-filter");
    TotalQuickFilter.setTextFilterBarButton(b, "qfb-text-filter");
    TotalQuickFilter.setHilightButton("textfiltercommand");
    TotalQuickFilter.setTextInputBox();
    TotalQuickFilter.initHilightButton();
    HiLighter.init();
    FolderDisplayListenerManager.registerListener(TotalQuickFilter.FolderDisplayListener);
//    MailServices.mailSession.AddFolderListener(TotalQuickFilter.FolderListener,
//                                               Ci.nsIFolderListener.propertyFlagChanged);
    TotalQuickFilter.FolderDisplayListener.onDisplayingFolder(gFolderDisplay);
    this.d("tabmail").registerTabMonitor(TotalQuickFilter.TabMonitor);
    this.d("mailContext").addEventListener("popupshowing", TotalQuickFilter.onMailContextPopupShowing);
    AddonManager.addAddonListener(TotalQuickFilter.AddonListener);
TotalQuickFilter.log.debug("init: END ");
  },

  orderedTextFilterDomIds: {},
  initTextButtons: function() {
    let p = this.d("quick-filter-bar-textfilter-bar");
    let x = this.d("quick-filter-bar-expando");
    if (!p || !x)
      return;

    let buttons = x.getElementsByTagName("toolbarbutton");
    let o = this.d("qfb-text-filter-popup");
    let filterValues = TotalQuickFilter.getTabFilterValues();
    let firstButton = true;
TotalQuickFilter.log.debug("initTextButtons: count - "+buttons.length);
    while (buttons.length) {
      let b = buttons[0];
      let name = b.getAttribute("name");
      if (name)
        b.label = TotalQuickFilter.searchStrings.GetStringFromName(name);
      name = b.getAttribute("displayname");
      if (name)
        b.setAttribute("displaylabel",
                       b.label + " (" +
                       TotalQuickFilter.searchStrings.GetStringFromName(name) + ")");
TotalQuickFilter.log.debug("initTextButtons: button - "+b.id);
      b.classList.add("toolbarbutton-1");
      b.classList.add("tqfButtonTextOnly");
      b.setAttribute("context", "");
      b.addEventListener("click",
        function(event) {
          TotalQuickFilter.clickFunc(event, this, null,
                                       TotalQuickFilter.checkTextFilterButton);
        });
      b.addEventListener("keypress",
        function(event) {
          TotalQuickFilter.keyFunc(event, this, null,
                                     TotalQuickFilter.checkTextFilterButton);
        });
      b.addEventListener("focus", TotalQuickFilter.focusHandler);
      b.addEventListener("mouseenter", TotalQuickFilter.focusHandler);
      p.appendChild(b);

      let a = document.createElement("arrowscrollbox");
      a.setAttribute("buttonid", b.id);
      a.setAttribute("orient", "horizontal");
      a.setAttribute("flex", "1");
      a.setAttribute("collapsed", true);
      p.appendChild(a);

      // Order of buttons must be maintained as that is how , delimited strings
      // are matched to the textfilter buttons.
      let implemented = !b.hasAttribute("notimplemented");
      if (implemented) {
        if (b.id in MessageTextFilter.textFilterDefsByDomId) {
          name = MessageTextFilter.textFilterDefsByDomId[b.id].name;
          TotalQuickFilter.orderedTextFilterDomIds[b.id] = name;
        }
        else {
          b.setAttribute("notimplemented", implemented = false);
TotalQuickFilter.log.info("initTextButtons: text filter button domId not " +
                            "found in textFilterDefs - "+b.id);
        }
      }

      let state = undefined;
      if (filterValues && filterValues.text)
        state = filterValues.text.states[name];
      let hidden = state === null || !implemented;

      // Boolean button.  First bool button is always AND and hidden.
      if (!firstButton) {
        let bb = document.createElement("toolbarbutton");
        bb.id = b.id + "-mode";
        bb.setAttribute("buttonid", b.id);
        bb.classList.add("toolbarbutton-1");
        bb.classList.add("tqfButtonIcon");
        bb.classList.add("tqfBooleanButton");
        bb.setAttribute("hidden", true);
        bb.setAttribute("value", "AND");
        bb.addEventListener("command", TotalQuickFilter.booleanButtonHandler);
        p.insertBefore(bb, b);
      }
      else
        firstButton = false;

      // Menuitem.
      let m = document.createElement("menuitem");
      m.id = b.id + "-menuitem";
      m.classList.add("tqf-textfilter-menuitem")
      m.setAttribute("type", "checkbox");
      m.setAttribute("label", b.getAttribute("displaylabel") || b.label);
      m.setAttribute("buttonid", b.id);
      if (!implemented)
        m.setAttribute("disabled", true);
      m.setAttribute("closemenu", "none");
//      m.setAttribute("oncommand",
//                     "TotalQuickFilter.selectSearchType(this, 'qfb-text-filter');");
      o.appendChild(m);

      TotalQuickFilter.setTextFilterButton(b, b.checked, null, hidden);
    }
  },

  initDateButtons: function() {
    let p = this.d("quick-filter-bar-datefilter-bar");
    let x = this.d("quick-filter-bar-datefilter-bar");
    if (!p || !x)
      return;

    let buttons = x.getElementsByTagName("toolbarbutton");
    let o = this.d("tqf-date-filter-popup");
    let filterValues = TotalQuickFilter.getTabFilterValues();
    let firstButton = true;
TotalQuickFilter.log.debug("initDateButtons: count - "+buttons.length);
    for (let b of buttons) {
      let name = b.getAttribute("name");
      if (name)
        b.label = TotalQuickFilter.searchStrings.GetStringFromName(name);
      name = b.getAttribute("displayname");
      if (name)
        b.setAttribute("displaylabel",
                       b.label + " (" +
                       TotalQuickFilter.searchStrings.GetStringFromName(name) + ")");
TotalQuickFilter.log.debug("initDateButtons: button - "+b.id);
      b.addEventListener("click",
        function(event) {
          TotalQuickFilter.clickFunc(event, this, null,
                                       TotalQuickFilter.checkTextFilterButton);
        });
      b.addEventListener("keypress",
        function(event) {
          TotalQuickFilter.keyFunc(event, this, null,
                                     TotalQuickFilter.checkTextFilterButton);
        });
      b.addEventListener("focus", TotalQuickFilter.focusHandler);
      b.addEventListener("mouseenter", TotalQuickFilter.focusHandler);

      let implemented = !b.hasAttribute("notimplemented");
      let state = undefined;
      if (filterValues && filterValues.text)
        state = filterValues.text.states[name];
      let hidden = state === null || !implemented;

      // Menuitem.
      let m = document.createElement("menuitem");
      m.id = b.id + "-menuitem";
      m.classList.add("tqf-textfilter-menuitem")
      m.setAttribute("type", "checkbox");
      m.setAttribute("label", b.getAttribute("displaylabel") || b.label);
      m.setAttribute("buttonid", b.id);
      if (!implemented)
        m.setAttribute("disabled", true);
      m.setAttribute("closemenu", "none");
//      m.setAttribute("oncommand",
//                     "TotalQuickFilter.selectSearchType(this, 'tqf-date-filter');");
      o.appendChild(m);

      TotalQuickFilter.setTextFilterButton(b, b.checked, null, hidden);
    }
  },

  initButtonsFunc: function(aClassName) {
TotalQuickFilter.log.debug("initButtonsFunc: aClassName - "+aClassName);
    let buttons = document.getElementsByClassName(aClassName);
    for (let b of buttons) {
TotalQuickFilter.log.debug("initButtonsFunc: button - "+b.id);
      b.setAttribute("context", "");
      b.addEventListener("click",
        function(event) {TotalQuickFilter.clickFunc(event, this, null, null);});
      b.addEventListener("keypress",
        function(event) {TotalQuickFilter.keyFunc(event, this, null, null);});
    }
  },

  initTextFilterButtonFunc: function(aId) {
    let f = this.d(aId);
    let b = f ? document.getAnonymousElementByAttribute(
                  f, "class", "box-inherit toolbarbutton-menubutton-button") :
                null;
    if (!b)
      return b;
TotalQuickFilter.log.debug("initTextFilterButtonFunc: aId:class - "+
                             aId+":"+b.className);
    b.checked = f.checked;
    b.classList.add("tqfTextFilterButtonIcon");
    b.classList.add("toolbarbutton-1");
    b.setAttribute("context", "");
    b.addEventListener("click",
      function(event) {TotalQuickFilter.clickFunc(event, this, null, null);});
    b.addEventListener("keypress",
      function(event) {TotalQuickFilter.keyFunc(event, this, null, null);});
    // Doesn't look like querySelector works here.
    let d = document.getAnonymousElementByAttribute(
              f, "class", "toolbarbutton-menubutton-dropmarker");
//    d.addEventListener("click", TotalQuickFilter.clickHandler);
    d.addEventListener("keypress",
      function(event) {TotalQuickFilter.keyFunc(event, this, null, null);});
    return b;
  },

  dispatchCommand: function(aTarget) {
TotalQuickFilter.log.debug("dispatchCommand: id - "+aTarget.id);
    setTimeout(function() { aTarget.dispatchEvent((new CustomEvent("command"))); }, 20);
  },

  dispatchHandler: function(aHandler, aEvent) {
TotalQuickFilter.log.debug("dispatchHandler:  ");
    setTimeout(function() { aHandler(aEvent); }, 20);
  },

  clickFunc: function(aEvent, aButton, aHandler, aRightClickHandler) {
TotalQuickFilter.log.debug("clickFunc: button id:className:checked - "+
                             aButton.id+":"+aButton.className+":"+
                             ("checked" in aButton ?
                                aButton.checked : "undefined"));
    if (!("button" in aEvent) || aButton.disabled)
      return;

    if (aEvent.button == 2) {
      aEvent.stopPropagation();
      aEvent.preventDefault();
      if (aRightClickHandler)
        this.dispatchHandler(aRightClickHandler, aEvent);
      else if (aButton.classList.contains("tqfRtClickInverts")) {
        aButton.checked = !aButton.checked;
        if (aButton.checked)
          aButton.setAttribute("inverted", true);
        else
          aButton.removeAttribute("inverted");
        this.dispatchCommand(aButton);
      }
      return;
    }
    else if (aEvent.button != 0)
      return;

    setTimeout(function() { TotalQuickFilter.finishEvent(aButton); }, 0);
  },

  keyFunc: function(aEvent, aButton, aHandler, aRightClickHandler) {
TotalQuickFilter.log.debug("keyFunc: button id:className - "+
                             aButton.id+":"+aButton.className);
    if (aEvent.which == KeyEvent.DOM_VK_SPACE) {
      aButton.checked = !aButton.checked;
      aButton.blur();
      return;
    }

    if (aButton.classList.contains("toolbarbutton-menubutton-dropmarker")) {
      if (aEvent.keyCode == KeyEvent.DOM_VK_RIGHT) {
        document.commandDispatcher.advanceFocus();
        return;
      }
      if (aEvent.keyCode == KeyEvent.DOM_VK_LEFT) {
        document.commandDispatcher.rewindFocus ();
        return;
      }
    }

    if (aEvent.which != KeyEvent.DOM_VK_RETURN)
      return;

    aEvent.stopPropagation();
    aEvent.preventDefault();

    if (aEvent.ctrlKey) {
TotalQuickFilter.log.debug("keyFunc: ctrlKey");
      if (aRightClickHandler)
        this.dispatchHandler(aRightClickHandler, aEvent);
      else if (aButton.classList.contains("tqfRtClickInverts")) {
        aButton.checked = !aButton.checked;
        if (aButton.checked)
          aButton.setAttribute("inverted", true);
        else
          aButton.removeAttribute("inverted");
        this.dispatchCommand(aButton);
      }
      return;
    }

    if ("type" in aButton && aButton.type == "checkbox")
      aButton.checked = !aButton.checked;

    this.finishEvent(aButton);

    if (aHandler)
      this.dispatchHandler(aHandler, aEvent);
    else
      this.dispatchCommand(aButton);
  },

  finishEvent: function(aButton) {
TotalQuickFilter.log.debug("finishEvent: button id:checked - "+
                             aButton.id+":"+
                             ("checked" in aButton ? aButton.checked : null));
    if (aButton.classList.contains("tqfRtClickInverts"))
      aButton.removeAttribute("inverted");

//    let t = this.d("qfb-qs-textbox");
    let p = aButton.parentNode;

    if (p.id == "tqf-date-filter") {
      if (aButton.classList.contains("toolbarbutton-menubutton-button")) {
        aButton.checked = !aButton.checked;
        TotalQuickFilter.setTextFilterBarButton(aButton, p.id);
//        TotalQuickFilter.setHilightButton("textfiltercommand");
      }
      if (aButton.classList.contains("toolbarbutton-menubutton-dropmarker"))
        this.d("tqf-date-filter-popup").showPopup();
    }

    if (p.id == "quick-filter-bar-datefilter-bar" &&
        aButton.classList.contains("tqfButtonTextOnly")) {
      // Make sure state is set asap.
//      TotalQuickFilter.setTabFilterValues(aButton, aButton.checked);
      TotalQuickFilter.setTextFilterBarButton(aButton, "tqf-date-filter");
//      TotalQuickFilter.setHilightButton("textfiltercommand");
    }

    if (p.id == "qfb-text-filter") {
      if (aButton.classList.contains("toolbarbutton-menubutton-button")) {
        aButton.checked = !aButton.checked;
        TotalQuickFilter.setTextFilterBarButton(aButton, p.id);
        TotalQuickFilter.setHilightButton("textfiltercommand");
      }
      if (aButton.classList.contains("toolbarbutton-menubutton-dropmarker"))
        this.d("qfb-text-filter-popup").showPopup();
    }

    if (p.id == "quick-filter-bar-textfilter-bar" &&
        aButton.classList.contains("tqfButtonTextOnly")) {
      // Make sure state is set asap.
      TotalQuickFilter.setTabFilterValues(aButton, aButton.checked);
      TotalQuickFilter.setTextFilterBarButton(aButton, "qfb-text-filter");
      TotalQuickFilter.setHilightButton("textfiltercommand");
    }

    if (p.tagName == "arrowscrollbox")
      setTimeout(function() { p.ensureElementIsVisible(aButton, false); }, 500);
  },

  focusHandler: function(event) {
    this.parentNode.ensureElementIsVisible(this, false);
    if (this.parentNode.id == "quick-filter-bar-textfilter-bar" && !this.disabled)
      this.parentNode.ensureElementIsVisible(this.nextElementSibling, false);
  },

  beforeCustomization: function() {
    let f = document.commandDispatcher.focusedElement;
    if (f instanceof HTMLInputElement)
      this.focusedElement = document.getBindingParent(f.parentNode);
    else
      this.focusedElement = f;
  },

  afterCustomization: function() {
    let t = TotalQuickFilter.d("qfb-qs-textbox");
    if (t.value) {
      t._searchIcons.selectedIndex = 1;
      t.setSelectionRange(t.textLength, t.textLength);
    }

    if (this.focusedElement)
      this.focusedElement.focus();

//    TotalQuickFilter.init();
    let b = TotalQuickFilter.initTextFilterButtonFunc("qfb-text-filter");
    TotalQuickFilter.setTextInputBox();
  },

  shouldUpdate: function(aButton) {
TotalQuickFilter.log.debug("shouldUpdate: aButton.id - "+aButton.id);
    if (aButton.parentNode.id != "quick-filter-bar-textfilter-bar" &&
        aButton.id != "qfb-qs-textbox")
      // Check is only for textfilter buttons or textinput.
      return true;

    let f = TotalQuickFilter.getTabFilterValues();
    let curTerms = f && f.text && f.text.curTerms ? f.text.curTerms : null;
    if (!curTerms)
      // Don't run a search if there's no text for the button.
      return false;

    let b = aButton.id == "qfb-qs-textbox" ? null : aButton;
    let p = aButton.id == "qfb-qs-textbox" ?
              this.d("quick-filter-bar-textfilter-bar") :
              aButton.parentNode;
    let [,,, actionButtons] = TotalQuickFilter.getCheckedChildren(b, p);
    let aButtonIndex = actionButtons.indexOf(aButton);
    if (aButtonIndex > curTerms.length - 1 ||
        (aButton.classList.contains("tqfBooleanButton") &&
         !aButton.nextElementSibling.nextElementSibling.childElementCount))
      return false;

    let index = 0;
    for (let a of actionButtons) {
      if (curTerms[index] && curTerms[index].length)
        return true;
      index++;
    }

    return false;
  },

/**
 * Return various states of the container aParentNode, or aButton's parentNode.
 *
 * @return [hasChild, firstCheckedChildId, nextCheckedChildId (after aButton),
 *           actionButtons (aButton and any other checked buttons)]
 */
  getCheckedChildren: function(aButton, aParentNode) {
    aParentNode = aParentNode ? aParentNode : aButton.parentNode;
    if (!aParentNode)
      return [false, 0, null, null, true];
    let buttons = aParentNode.getElementsByTagName("toolbarbutton");
    let visIds = [];
    let chkIds = [];
    let actionButtons = []

    for (let b of buttons) {
      if (b.tagName != "toolbarbutton")
        continue;
      if (!b.hidden && b.parentNode == aParentNode)
        visIds.push(b.id);
      if (b.checked && b.parentNode == aParentNode)
        chkIds.push(b.id);
      if ((aButton && b.id == aButton.id && b.parentNode == aParentNode) ||
          (aButton && b.id != aButton.id && b.checked && b.parentNode == aParentNode) ||
          (aButton === null && b.checked && b.parentNode == aParentNode))
        actionButtons.push(b);
    }
    let i = aButton ? chkIds.indexOf(aButton.id) : -1;
    let nextChkId = i >= 0 && chkIds.length >= i + 2 ? chkIds[i + 1] : null;
TotalQuickFilter.log.trace("getCheckedChildren: aParent - "+aParentNode.id+" : "+
                             [visIds.length ? true : false,
                              chkIds.length ? chkIds[0] : null,
                              nextChkId,
                              actionButtons]);

    return [visIds.length ? true : false,
            chkIds.length ? chkIds[0] : null,
            nextChkId,
            actionButtons];
  },

  isHeaderChecked: function() {
    let parent = this.d("quick-filter-bar-textfilter-bar");
    if (!parent)
      return false;

    let buttons = parent.getElementsByTagName("toolbarbutton");
    for (let b of buttons) {
      if (b.checked && b.parentNode == parent && b.id != "qfb-qs-body")
        return true;
    }

    return false;
  },

  isBodyChecked: function() {
    let b = this.d("qfb-qs-body");
    return b && b.checked ? b : false;
  },

  isHilighterStateActive: function() {
    let f = this.getTabFilterValues();
    return f && f.hilighter ? true : false;
  },

  isInFindMode: function() {
    return !this.isHeaderChecked() && !this.isBodyChecked();
  },

  getButtonsBar: function() {
    let b = this.isBodyChecked();
    return b ? b.nextElementSibling : this.d("quick-filter-bar-terms-bar");
  },

  getTabFilterValues: function(aFolderDisplay) {
    let folderDisplay = aFolderDisplay ? aFolderDisplay : gFolderDisplay;
    return ("quickFilter" in folderDisplay._tabInfo._ext) ?
             folderDisplay._tabInfo._ext.quickFilter.filterValues : null;
  },

  setTabFilterValues: function(aButton, aValue) {
    let f = TotalQuickFilter.getTabFilterValues();
    let name = TotalQuickFilter.orderedTextFilterDomIds[aButton.id];
    if (f && f.text) {
      // Use null for per tab hidden button state, persisted in session.json.
      f.text.states[name] = aValue;
      return true;
    }
    return false;
  },

  getModes: function(aFilterValue) {
TotalQuickFilter.log.trace("getModes: modes - "+ (aFilterValue.modes ?
                             aFilterValue.modes.toSource() : null));
    if (aFilterValue.modes)
      return aFilterValue.modes;
    return aFilterValue.modes = {};
  },

  booleanButtonHandler: function(aEvent) {
    let b = aEvent.target;
    // Toggle.
    let val = b.value == "OR" ? true : false;
    TotalQuickFilter.setBooleanButton(b, val);
    let filterValue =
        QuickFilterBarMuxer.getFilterValueForMutation(MessageTextFilter.name);
    let domId = aEvent.target.getAttribute("buttonid");
    let name;
    if (domId in MessageTextFilter.textFilterDefsByDomId)
      name = MessageTextFilter.textFilterDefsByDomId[domId].name;
    if (name)
      // Modes textfilter by name value is false if boolean OR, defaults to AND.
      TotalQuickFilter.getModes(filterValue)[name] = val;

    if (TotalQuickFilter.shouldUpdate(b))
      QuickFilterBarMuxer.updateSearch();
  },

  setBooleanButton: function(aButton, aVal) {
TotalQuickFilter.log.debug("setBooleanButton: buttonId:aVal - "+
                             aButton.id+":"+aVal);
    let bb = this.d("tqf-boolean-mode-textfilter-begingroup");
    let newVal = aVal ? "AND" : "OR";
    let label = aVal ? "labeland" : "labelor";
    aButton.setAttribute("value", aButton.value = newVal);
    aButton.setAttribute("label", bb.getAttribute(label));
    aButton.setAttribute("tooltiptext", bb.getAttribute(label));
  },

  setTagBooleanButton: function(aVal) {
TotalQuickFilter.log.debug("setTagBooleanButton: aVal - "+aVal);
    let bb = this.d("tqf-boolean-mode");
    let val = aVal === null ? bb.getAttribute("value") == "AND" ? true : false : aVal;
    let newVal = val ? "AND" : "OR";
    let label = val ? "labelall" : "labelany";
    let tooltip = val ? "tooltipall" : "tooltipany";
    bb.setAttribute("value", bb.value = newVal);
    bb.setAttribute("label", bb.getAttribute(label));
    bb.setAttribute("tooltiptext", bb.getAttribute(tooltip));
  },

  setTextFilterBarButton: function(aButton, aId) {
    let b = this.d(aId);
    if (!b || !aButton)
      return;

TotalQuickFilter.log.debug("setTextFilterBarButton: aId:buttonId:checked - "+
                             aId+":"+aButton.id+":"+
                             ("checked" in aButton ? aButton.checked : null));
    let pId = aId == "tqf-date-filter" ? "quick-filter-bar-datefilter-bar" :
              aId == "qfb-text-filter" ? "quick-filter-bar-textfilter-bar" :
                                         null;
    if (!pId)
      return;
    let parent = this.d(pId);
    let [hasChild, firstCheckedChildId,,] =
        TotalQuickFilter.getCheckedChildren(aButton, parent);
    if (hasChild)
      parent.setAttribute("haschild", true);
    else
      parent.removeAttribute("haschild");
    if (firstCheckedChildId)
      parent.setAttribute("checkedchild", true);
    else
      parent.removeAttribute("checkedchild");

    if (b == aButton.parentNode)
      b.checked = aButton.checked;
    if (b.checked && hasChild)
      parent.removeAttribute("collapsed");
    else
      parent.setAttribute("collapsed", true);
  },

  setTextFilterPopup: function(aPopup) {
    for (let m of aPopup.childNodes) {
      if (m.tagName == "menuitem") {
        let b = this.d(m.getAttribute("buttonid"));
        if (!b || b.hasAttribute("notimplemented"))
          continue;

        m.disabled = b.disabled;
        if (b.checked)
          m.setAttribute("selected", true);
        else
          m.removeAttribute("selected");
        if (!b.hidden)
          m.setAttribute("checked", true);
        else
          m.removeAttribute("checked");
      }
    }
  },

  selectSearchType: function(aMenuItem, aButtonId) {
    let b = this.d(aMenuItem.getAttribute("buttonid"));
    if (!b)
      return;

    let checked = aMenuItem.getAttribute("checked");
    let selected = aMenuItem.getAttribute("selected");
    if (selected)
      aMenuItem.removeAttribute("selected");
    if (!checked && !selected) {
      aMenuItem.setAttribute("checked", true);
      aMenuItem.setAttribute("selected", true);
    }

    let state = !aMenuItem.getAttribute("checked")           ? null :
                aMenuItem.getAttribute("selected") == "true" ? true :
                                                               false;
    TotalQuickFilter.setTabFilterValues(b, state);
    let t = this.d(aButtonId);
    TotalQuickFilter.setTextFilterBarButton(t, t.id);
    TotalQuickFilter.setHilightButton("textfiltercommand");
    if ((selected || !checked) && TotalQuickFilter.shouldUpdate(b))
      QuickFilterBarMuxer.updateSearch();
  },

  checkTextFilterButton: function(aEvent) {
    let b = aEvent.target;
    if (b.disabled)
      return;

    TotalQuickFilter.doCheckTextFilterButton(b, true);
    TotalQuickFilter.setHilightButton("textfiltercommand");
    TotalQuickFilter.dispatchCommand(b);
  },

  doCheckTextFilterButton: function(aButton, aUncheckAll) {
    // Check only the selected button, uncheck the rest.  Uncheck all if
    // selected button is checked and aUncheckAll is true.
    for (let button of aButton.parentNode.children) {
      if (button.tagName == "toolbarbutton" && !button.disabled &&
          (button.checked || button.id == aButton.id)) {
        let checked = aButton.checked && aUncheckAll ?
                        false : button.id == aButton.id;
        button.checked = checked;
        TotalQuickFilter.setTabFilterValues(button, checked);
      }
    }

    TotalQuickFilter.setTextFilterBarButton(aButton, "qfb-text-filter");
  },

  setTextFilterButton: function(aButton, aChecked, aDisabled, aHidden) {
    if (!aButton)
      return;

    aButton.hidden = aHidden === null ? false : aHidden;
    let t = this.d("qfb-text-filter");
    if (aDisabled === true || (aDisabled === null && aButton.disabled)) {
      aButton.disabled = true;
      if (t)
        aButton.setAttribute("tooltiptext",
                             t.getAttribute("notextfiltertooltip"));
    }
    else {
      aButton.disabled = false;
      aChecked = aChecked === null ? aButton.checked : aChecked;
    }
    if (aChecked === true) {
      aButton.checked = true;
      if (!aButton.disabled)
        aButton.removeAttribute("tooltiptext");
    }
    if (aChecked === false) {
      aButton.checked = false;
      if (t && !aButton.disabled)
        aButton.setAttribute("tooltiptext",
                             t.getAttribute("uncheckedtextfiltertooltip"));
    }
    aButton.checkState = aButton.checked ? 1 : 0;

    // Textfilter button's boolean button.
    let bb = this.d(aButton.id + "-mode");
    let [, firstCheckedChildId, nextCheckedChildId,] =
        TotalQuickFilter.getCheckedChildren(aButton);
    let first = this.d(firstCheckedChildId + "-mode");
    let next = this.d(nextCheckedChildId + "-mode");
TotalQuickFilter.log.trace("setTextFilterButton: b:bb:first:next:" +
                             "checked:disabled:hidden - "+
                             aButton.id+":"+(bb ? bb.id : null)+":"+
                             firstCheckedChildId+":"+nextCheckedChildId+":"+
                             aChecked+":"+aDisabled+":"+aHidden);
    if (aButton.checked) {
      if (bb && aButton.id != firstCheckedChildId)
        bb.removeAttribute("hidden");
      if (next)
        next.removeAttribute("hidden");
    }
    else {
      if (bb)
        bb.setAttribute("hidden", true);
      if (first)
        first.setAttribute("hidden", true);
      if (next)
        next.setAttribute("hidden", true);
    }
  
  },

  initHilightButton: function() {
    // This button has a custom binding.
    let b = this.d("tqf-terms");
    if (!b)
      return;

    let rightClickHandler = function(aEvent) {
      let b = TotalQuickFilter.d("tqf-terms");
      let m = TotalQuickFilter.d("tqf-terms-markers");
      let active = b.getAttribute("markers") != "true";
      b.setAttribute("markers", active);
      HiLighter._HilightMarkersMode = null;
      if (active) {
        m.setAttribute("style", "background-color: " +
                                HiLighter.MarkersBackgroundPref);
        HiLighter.doHilightMarkersAll(HiLighter.highlightBrowserElement);
      }
      else {
        m.removeAttribute("style");
        HiLighter.hilightMarkersClear(HiLighter.highlightBrowserElement);
      }
    }

    b.setAttribute("context", "");
    b.addEventListener("click",
      function(event) {
        TotalQuickFilter.clickFunc(event, this, null, rightClickHandler);});
    b.addEventListener("keypress",
      function(event) {
        TotalQuickFilter.keyFunc(event, this, null, rightClickHandler);});

    let m = this.d("tqf-terms-markers");
    if (b.getAttribute("markers") == "true" && m)
      m.setAttribute("style", "background-color: " +
                              HiLighter.MarkersBackgroundPref);
TotalQuickFilter.log.debug("initHilightButton: ");
  },

  setHilightButton: function(aCommand) {
    let b = this.d("tqf-terms");
    let parent = this.getButtonsBar();
    if (!b || !parent)
      return;

    b.removeAttribute("busy");
    let active = b.checked;
    let hadChild = parent.hasAttribute("haschild");
    let filterValues = this.getTabFilterValues();
    let findMode = this.isInFindMode();
TotalQuickFilter.log.debug("setHilightButton: aCommand:checked:findMode - "+
                             aCommand+":"+active+":"+findMode);
    if (filterValues && filterValues.text) {
      if (findMode) {
        this.setTermsButton(filterValues.text.curTerms ?
                              filterValues.text.curTerms[0] : [],
                            "tqf-terms", active);
        if (!active)
          parent.setAttribute("collapsed", true);
      }
      else {
        this.setTermsButton([], "tqf-terms", active);
      }

      // Set the textfilter buttons.
      this.setTermsButtons(filterValues.text);
    }

    let hasChild = parent.hasAttribute("haschild");
    if (active) {
      // For findmode, this is the fastest way to clear old spans if any item
      // in the text search changes.
      if (gFolderDisplay.selectedCount) { // == 1) {
        if (aCommand == "textfiltercommand")
          ReloadMessage();
        else if (hasChild)
          HiLighter.doHilight(aCommand, null, HiLighter.highlightBrowserElement);
      }
    }
    else {
      if (hadChild && gFolderDisplay.selectedCount)
        HiLighter.hilightHide(null, HiLighter.highlightBrowserElement);
    }
  },

  setTermsButtons: function(aTextFilterValue) {
    if (!aTextFilterValue) {
      let filterValues = this.getTabFilterValues();
      aTextFilterValue = filterValues && filterValues.text ?
                           filterValues.text : null;
    }
    if (!aTextFilterValue)
      return;
TotalQuickFilter.log.debug("setTermsButtons: START aTextFilterValue - "+
                             aTextFilterValue.toSource());
    let i = 0;
    for (let [tfDomId, tfName] in Iterator(TotalQuickFilter.orderedTextFilterDomIds)) {
      let terms = [];
      let b = this.d(tfDomId);
      let state = aTextFilterValue.states[tfName];
      let mode = this.getModes(aTextFilterValue)[tfName];
      if (state && aTextFilterValue.curTerms)
        // Button must be active and there must be terms for it, otherwise make
        // sure it's empty (eg when switching tabs).
        terms = aTextFilterValue.curTerms[i++] || [];

TotalQuickFilter.log.debug("setTermsButtons: state:tfName:domId:terms - "+
                             state+":"+tfName+":"+tfDomId+":"+terms);
      this.setTextFilterButton(b, state === true, null, state === null);
      let bb = this.d(tfDomId + "-mode");
      if (bb)
        this.setBooleanButton(bb, mode === false ? false : true);
      this.setTermsButton(terms, tfDomId, this.isHilighterStateActive());
    }
  },

  setTermsButton: function(aNewTerms, aButtonId, aHilighterActive) {
    let termButton = this.d(aButtonId);
    if (!termButton)
      return;
    let hlTerm = aHilighterActive &&
                 (aButtonId == "qfb-qs-body" || aButtonId == "tqf-terms");
    let termsBar = aButtonId == "tqf-terms" ?
                     this.d("quick-filter-bar-terms-bar") :
                     termButton.nextElementSibling;
    if (!termsBar)
      return;
TotalQuickFilter.log.trace("setTermsButton: newTerms:hlTerm - "+
                             aNewTerms+":"+hlTerm);

    let buttonTerms = [];
    if (!termButton.hidden)
      buttonTerms = aNewTerms.filter(function(term) {
                                       return term != TotalQuickFilter.OR;
                                     });
    buttonTerms.forEach(function(term, i) {
      let button = this.d(aButtonId + "-term-" + i);
      let [negated, or, quoted, phrase] = this.parseTerm(term);
/*      if (button && button.getAttribute("displayterm") == term) {
TotalQuickFilter.log.debug("setTermsButtons: No change i:term - "+i+":"+term);
        button.label = button.getAttribute("displayterm");
        if (hlTerm && !negated) {
          let style = HiLighter.getRuleString(button, HiLighter.getRuleIndex(i));
          button.setAttribute("style", style);
        }
//        if (hlTerm)
//          button.classList.add("tqfHlButton");
//        else
//          button.classList.remove("tqfHlButton");
        return;
      }*/
      if (!button) {
        // Create.
TotalQuickFilter.log.trace("setTermsButton: Create i:term - "+i+":"+term);
        button = document.createElement("toolbarbutton");
        button.setAttribute("id", aButtonId + "-term-" + i);
//        button.classList.add("toolbarbutton-1");
        button.classList.add("tqfButtonTextOnly");
        button.classList.add("tqfTermButton");
        button.classList.add("tqfRtClickInverts");
        button.classList.add(button.id);
        button.setAttribute("index", i);
        button.setAttribute("context", "");
        button.addEventListener("command", TotalQuickFilter.termsHandler);
        button.addEventListener("click",
          function(event) {
            TotalQuickFilter.clickFunc(event, this,
                                         TotalQuickFilter.termsHandler,
                                         TotalQuickFilter.termsRightClickHandler);
          });
        button.addEventListener("keypress",
          function(event) {
            TotalQuickFilter.keyFunc(event, this,
                                       TotalQuickFilter.termsHandler,
                                       TotalQuickFilter.termsRightClickHandler);
          });
        button.addEventListener("focus", TotalQuickFilter.focusHandler);
        button.addEventListener("mouseenter", TotalQuickFilter.focusHandler);
        termsBar.appendChild(button);
      }
      else {
        // Update.
TotalQuickFilter.log.trace("setTermsButton: Update i:term - "+i+":"+term);
        if (button.getAttribute("displayterm") != term)
          button.removeAttribute("hilit");
      }

      if (negated) {
        button.setAttribute("type", "button");
        button.removeAttribute("checked");
        button.setAttribute("negated", true);
        button.removeAttribute("style");
      }
      else {
        button.removeAttribute("negated");
        if (hlTerm) {
          button.classList.add("tqfHlButton");
          button.setAttribute("type", "checkbox");
          button.setAttribute("checked", true);
          let style = HiLighter.getRuleString(button, HiLighter.getRuleIndex(i));
          button.setAttribute("style", style);
        }
        else {
          button.classList.remove("tqfHlButton");
          button.setAttribute("type", "button");
          button.removeAttribute("checked");
          button.removeAttribute("style");
        }
      }
      if (quoted)
        button.setAttribute("quoted", true);
      else
        button.removeAttribute("quoted");
      button.setAttribute("label", phrase);
      button.setAttribute("displayterm", phrase);
      button.setAttribute("term", this.IgnoreCase ? phrase.toLowerCase() : phrase);
    }, this);

    while (buttonTerms.length < termsBar.childElementCount)
      termsBar.lastChild.remove();

    if (termsBar.childElementCount) {
      termsBar.setAttribute("haschild", true);
      termsBar.removeAttribute("collapsed");
    }
    else {
      termsBar.setAttribute("collapsed", true);
      termsBar.removeAttribute("haschild");
    }
  },

  termsHandler: function(aEvent) {
    let button = aEvent.target;
    if (button.hasAttribute("negated"))
      TotalQuickFilter.termsRightClickHandler(aEvent);
    else if (button.classList.contains("tqfHlButton"))
      HiLighter.doHilight("button", button, HiLighter.highlightBrowserElement);
  },

  termsRightClickHandler: function(aEvent) {
TotalQuickFilter.log.debug("termsRightClickHandler: button - "+
                             aEvent.target.label);
    let t = TotalQuickFilter.d("qfb-qs-textbox");
    let f = TotalQuickFilter.getTabFilterValues();
    if (!t || !f || !f.text)
      return;

    let b = aEvent.target;
    let n = b.hasAttribute("negated");
    let r = TotalQuickFilter.orderedTextFilterDomIds;
    let fmode = b.parentNode.id == "quick-filter-bar-terms-bar";

    if (fmode) {
      // In findmode.
      if (!n)
        // No negation in findmode.
        return;
      // Turn negated button in terms bar into hilight button.
      b.removeAttribute("negated");
      r = {"tqf-terms":"hilighter"};
    }

    let d = b.getAttribute("displayterm");
    let q = b.hasAttribute("quoted");
    let v = [];
    let k = 0
    for (let [, m] in Iterator(r)) {
      if (m != "hilighter" && !f.text.states[m])
        continue;
      let p = [].concat(f.text.curTerms[k++] || []);
      let s = d.replace(/"/g,'\\"');
      s = q ? '"' + s + '"' : s;
      s = n ? "-" + s : s;
      let j = p.indexOf(s);
      if (j != -1 && r[b.parentNode.previousElementSibling.id] == m) {
        p[j] = n ? p[j].replace(/^-/,"") : p[j] = "-" + p[j];
        if (fmode)
          f.text.curTerms[0][j] = p[j];
      }
      v.push(p.join(" ").replace(/,/g,'\\,'));
    }
    while (k < f.text.curTerms.length)
      v.push(([].concat(f.text.curTerms[k++] || [])).join(" ").replace(/,/g,'\\,'));
    while (v[v.length - 1] == "")
      v.pop();
    t.value = v.join(",");
    if (fmode) {
      f.text.text = t.value;
      TotalQuickFilter.setHilightButton("rtclick");
    }
    else {
      TotalQuickFilter.dispatchCommand(t);
    }

    setTimeout(function() { b.focus(); }, 40);
  },

  setTextInput: function(aNode) {
TotalQuickFilter.log.debug("setTextInput: "+aNode.id);
    // Update hilight buttons, if active.
    let h = this.d("tqf-terms");
    if (h && h.checked)
      TotalQuickFilter.setHilightButton("textfiltercommand");
    else
      TotalQuickFilter.setTermsButtons();

    aNode.setAttribute("value", aNode.value);
  },

  get SearchDelayPref() {
    return Services.prefs.getIntPref(this.prefBranch + ".searchDelayOnIput");
  },

  setTextInputBox: function() {
    let t = this.d("qfb-qs-textbox");
    if (!t)
      return;

    // Set the keyinput delay pref.
    t.addEventListener("keypress",
      function(event) {
        // Immediate feedback.
//        if (event.charCode) {
//          this.d("quick-filter-bar").setAttribute("filterActive", "searching");
//        }
        this.setAttribute("timeout", TotalQuickFilter.SearchDelayPref);
      });

    // Set the clear button.
    let c = document.getAnonymousElementByAttribute(t, "class",
                                                    "textbox-search-clear");
    function commandHandler() {document.getBindingParent(c)._clearSearch();};
    c.addEventListener("keypress",
      function(event) {TotalQuickFilter.keyFunc(event, this, commandHandler);});
TotalQuickFilter.log.debug("setTextInputBox: "+c.className);
  },

  get tqfBundle() {
    return this._tqfBundle ? this._tqfBundle : this._tqfBundle =
      Services.strings.createBundle("chrome://totalquickfilter/locale/tqf.properties");
  },

  onMailContextPopupShowing: function(event) {
    if (event.target != event.currentTarget)
      return;

    let selection = document.commandDispatcher.focusedWindow.getSelection();
    let isContentSelected = !selection.isCollapsed;
    let isMessagePane = document.popupNode.ownerGlobal.name == "messagepane";
    let show = isContentSelected && isMessagePane;
    for (let menuitemId of ["mailContext-addToQuickFilter",
                            "mailContext-replaceInQuickFilter"]) {
      let menuItem = TotalQuickFilter.d(menuitemId);
      if (show) {
        let term = selection.toString();
        let label = term.length < 15 ? "mailContext-QuickFilterMenuitem" :
                                       "mailContext-QuickFilterMenuitemE";
        let mi = TotalQuickFilter.tqfBundle.GetStringFromName(menuitemId);
        term = term.length < 15 ? term : term.slice(0, 14);

        menuItem.label =
          TotalQuickFilter.tqfBundle.formatStringFromName(label, [mi, term], 2);
        menuItem.value = term;
//        menuItem.setAttribute("term", term);
        menuItem.removeAttribute("hidden");
      }
      else {
        menuItem.setAttribute("hidden", true);
      }
    }
TotalQuickFilter.log.debug("onMailContextPopupShowing: DONE");
  },

  /**
   *  Terms processing.
   */
  IgnoreCase: true,
  // Unquoted space surrounded | means OR.  We do not use OR because it is
  // 1)twice the typing, 2)english-centric.
  // Note: 'the| one' means 'the|' AND 'one'; 'the | one' means 'the' OR 'one'.
  OR: "|",
  getCurTerms: function(aFilterValue) {
TotalQuickFilter.log.debug("getCurTerms: curTerms - "+aFilterValue.curTerms);
    if (aFilterValue.curTerms)
      return aFilterValue.curTerms;
    return aFilterValue.curTerms = this.onFilter(aFilterValue.text, aFilterValue);
  },

  get MinTermLengthPref() {
    return Services.prefs.getIntPref(this.prefBranch + ".minTermLength");
  },

  parseTerm: function(aTerm) {
    let term = aTerm;
    let negated = term.startsWith("-");
    let or = term == this.OR;
    let quoted = term.match(/^"|[^\\]"/g);
    // Handling to 1)remove negation -, 2)remove leading quote ",
    // 3)remove last quote ", 4)unescape escaped quotes \".
    term = term.replace(/^-/, "")
               .replace(/^"/, "")
               .replace(/([^\\]{1})"$/, "$1")
               .replace(/\\"/g, '"');

    return [negated, or, quoted, term];
  },

  clearFilter: function() {
    let t = this.d("qfb-qs-textbox");
    (t && t.removeAttribute("invalid"))
  },

  onFilterError: function() {
    let t = this.d("qfb-qs-textbox");
    (t && t.setAttribute("invalid", true))
    return false;
  },

  onFilter: function(aText, aFilterValue) {
    this.clearFilter();
    if (aText === null)
      return null;

TotalQuickFilter.log.debug("onFilter: text string - "+aText);
    // Negative lookbehinds, meh.
    let textStr = aText.replace(/\\,/g, "\x01");
    // Don't start with space or |.
    if (textStr.startsWith(" ") || textStr.startsWith(this.OR))
      return this.onFilterError();
    if (textStr.match(/\s\|$|[,\s]$/))
      // Don't run search on a delimiting , or a space.
      return false;

    // Create a grouped terms array from the input.  Each item is delimited by
    // a , and corresponds to a checked textfilter button.  Searches containing
    // , must be escaped with \.
    let textFilterGroups = [];
    let groups = textStr.split(/,/);
    while (groups.length) {
      let group = groups.shift();
TotalQuickFilter.log.debug("onFilter: text group - "+ group);
      let phrases = this.textFilterGroup(group);
      if (phrases === false)
        return false;
      textFilterGroups.push(phrases)
    }
TotalQuickFilter.log.debug("onFilter: END textFilterGroups - "+
                             textFilterGroups);
    return textFilterGroups;
  },

  textFilterGroup: function(aValue) {
    // Edit check routine for a textfilter group of search terms.
    if (!aValue)
      return [];

    // Restore escaped , from above.
    let value = aValue.replace(/\x01/g, ",");
    let quoted = value.match(/^"|[^\\]"/g);
    let quotesClosed = quoted && quoted.length%2 == 0;
    if (quotesClosed && value.match(/[^\\]"\S$/))
      return this.onFilterError();

TotalQuickFilter.log.debug("textFilterGroup: value - "+ value);
    let phrases = [];
    // Create a terms array from the input. A - indicates a negation term; the
    // item immediately following - may be a quoted string. A | indicates OR.
    let terms = value.match(/(-?"(?:[^\\"]+|\\+.)*?"|\S+)/g);
    let count = terms.length;
TotalQuickFilter.log.debug("textFilterGroup: terms - "+ terms);

    while (terms.length) {
      let term = terms.shift();
//      term = this.IgnoreCase ? term.toLowerCase() : term;
TotalQuickFilter.log.debug("textFilterGroup: term - "+term);

      // Term: quoted term must be non empty.
      if (term.match(/^""|-""/))
        return this.onFilterError();

      let testerm = term.replace(/^-/, "");
      let quoted = testerm.match(/^"|[^\\]"/g);
      let quotesClosed = quoted && quoted.length%2 == 0;
      // Term: opening quote " can only be preceded by a single negation - and
      // a closing quote must be followed by a space.  Otherwise searches
      // containing " need to be escaped with \.  Wait for a quoted term to
      // close before proceeding.
TotalQuickFilter.log.debug("textFilterGroup: testerm - "+testerm);
      if (quoted) {
        if (!quotesClosed) {
          if (testerm.match(/^[^"]/))
            return this.onFilterError();
          else
            return false;
        }
        else {
          if (testerm.match(/[^"]$/))
            return this.onFilterError();
        }
      }

      // Term: fts NEAR ~ invalid.
//      if (term.startsWith("~"))
//        return this.onFilterError();

      if (term == "-") {
        if (terms.length || (value.endsWith(" ") && !terms.length))
          return this.onFilterError();
      }
      else if (term == this.OR)
        phrases.push(term);
      else {
        // Term: unquoted term length too short.  If not the last term, or the
        // last term followed by a space, set error.
        let checkLen = term.startsWith("-") ?
                       this.MinTermLengthPref + 1 : this.MinTermLengthPref;
        if (term.length < checkLen) {
          if (terms.length || (value.endsWith(" ") && !terms.length))
            return this.onFilterError();
          return false;
        }

        let nodupe;
        if (this.IgnoreCase)
          nodupe = phrases.every(function(p) {
                                   return p.toLowerCase() != term.toLowerCase()
                                 });
        else
          nodupe = phrases.indexOf(term) == -1;

        if (nodupe)
          // No dupes.
          phrases.push(term);
      }
    }

    this.clearFilter();
TotalQuickFilter.log.debug("textFilterGroup: searchRE phrases - "+ phrases);
    return phrases;
  },

  searchForThis: function(aEvent) {
    aEvent.stopPropagation();
    if (!gFolderDisplay || !gFolderDisplay.selectedCount)
      return false;

    let msgHdr = gFolderDisplay.selectedMessage;
    let searchTerm = null;
    let tfName = aEvent.target.value;
    let searchBy = aEvent.target.getAttribute("searchby") == "name" ? "name" :
                                                                      "email";
TotalQuickFilter.log.debug("searchForThis: "+ tfName+ " "+searchBy);

    let createEmailTerm = function(aVal) {
      let emails = {}; let names = {}; let termsArr = [];
      MailServices.headerParser.parseHeadersWithArray(aVal, emails, names, {});
TotalQuickFilter.log.debug("searchForThis: emails:names = "+
                           emails.value+" : "+names.value);
      if (!emails.value.toString())
        searchBy = "name";

      let terms = searchBy == "name" ? names.value.toString().split(",") :
                                       emails.value.toString().split(",");
      let quote = searchBy == "name" ? '"' : "";
      for (let term of terms) {
        quote = term ? quote : '"';
        termsArr.push(quote + term.replace(/("|,)/g, "\\$1") + quote);
      }

      return searchTerm = termsArr.join(" | ");
    }

    switch (tfName) {
      case "subject":
        searchTerm = msgHdr.mime2DecodedSubject.replace(/^(Re:|fwd:|fw:)*/i, "");
        searchTerm = '"' + searchTerm.trim().replace(/("|,)/g, "\\$1") + '"';
        break;
      case "sender":
        searchTerm = createEmailTerm(msgHdr.mime2DecodedAuthor);
        break;
      case "recipients":
        searchTerm = createEmailTerm(msgHdr.mime2DecodedRecipients);
        break;
      case "to":
        searchTerm = createEmailTerm(msgHdr.mime2DecodedTo);
        break;
      case "cc":
        searchTerm = createEmailTerm(msgHdr.mime2DecodedCc);
        break;
      case "bcc":
        searchTerm = createEmailTerm(msgHdr.mime2DecodedBcc);
        break;
      default:
        break;
    }

    return this.searchFor(searchTerm, tfName)
  },

  searchForThisEmail: function(aEvent) {
    let node = findEmailNodeFromPopupNode(document.popupNode, "emailAddressPopup");
    let tfName = aEvent.target.value;
TotalQuickFilter.log.debug("searchForThisEmail: "+ tfName);
    return this.searchFor(node.getAttribute("emailAddress"), tfName);
  },

  searchForThisText: function(aEvent, aAdd) {
    let term = document.commandDispatcher.focusedWindow.getSelection().toString();
    let quote = term.match(/\s/) ? '"' : "";
    term = quote + term.replace(/("|,)/g, "\\$1") + quote;
TotalQuickFilter.log.debug("searchForThisText: "+ term);
    let filterTextBox = this.d("qfb-qs-textbox");
    let filterTextBar = this.d("quick-filter-bar-textfilter-bar");
    if (!term || !filterTextBox || !filterTextBar)
      return false;
    goDoCommand("cmd_showQuickFilterBar");
    filterTextBox.value = aAdd && filterTextBox.value ?
                            filterTextBox.value + " " + term : term;
    let f = TotalQuickFilter.getTabFilterValues();
    (f && (f.text.curTerms = f.text.text = null))
    // Immediate feedback.
    this.d("quick-filter-bar").setAttribute("filterActive","searching");
    filterTextBox.doCommand();
    setTimeout(function() { getBrowser().focus(); }, 40);
    return true;
  },

  searchFor: function(aSearchTerm, aTextFilterName) {
    let filterTextBox = this.d("qfb-qs-textbox");
    let filterTextBar = this.d("quick-filter-bar-textfilter-bar");
    let tfDef = MessageTextFilter.textFilterDefs[aTextFilterName];
    let domId = tfDef ? tfDef.domId : null;
    let button = this.d(domId);
    if (aSearchTerm === null || !button || !filterTextBox || !filterTextBar)
      return false;
TotalQuickFilter.log.debug("searchFor: "+ aSearchTerm);

    goDoCommand("cmd_showQuickFilterBar");
    filterTextBox.value = aSearchTerm;
    button.checked = true;
    let f = TotalQuickFilter.getTabFilterValues();
    (f && (f.text.curTerms = f.text.text = null))
    TotalQuickFilter.doCheckTextFilterButton(button, false);
    filterTextBox.doCommand();
    setTimeout(function() { GetThreadTree().focus(); }, 40);
    return true;
  },

  setShowFilterPopup: function(aPopup) {
    let disabled = !gFolderDisplay || !gFolderDisplay.selectedCount;
    for (let m of aPopup.getElementsByAttribute("value", "*")) {
      m.disabled = disabled;
      m.removeAttribute("checked");
    }
  },

  QuickFilterInfo: function(aButton) {
    let p = this.d("quickFilterInfo");
    let top = aButton.boxObject.y + aButton.parentNode.clientHeight * 1.5 + "px; ";
    let left = (aButton.boxObject.x - p.clientWidth / 2) + "px;";
    if (aButton.checked)
      p.setAttribute("style", "top: " + top + "left: " + left);
    else
      p.setAttribute("style", "left: " + left);
  },

//  FolderListener: {
//    OnItemAdded: function(parentItem, item) { },
//    OnItemRemoved: function(parentItem, item) { },
//    OnItemPropertyChanged: function(item, property, oldValue, newValue) { },
//    OnItemIntPropertyChanged: function(item, property, oldValue, newValue) { },
//    OnItemBoolPropertyChanged: function(item, property, oldValue, newValue) { },
//    OnItemUnicharPropertyChanged: function(item, property, oldValue, newValue) { },
//    OnItemEvent: function(folder, event) { },
//    OnItemPropertyFlagChanged: function(item, property, oldFlag, newFlag) {
      // Get the msgLoaded notification. The load event is too early.
//      if (property.toString() != "msgLoaded")
//        return;
//TotalQuickFilter.log.debug("OnItemPropertyFlagChanged: "+property.toString());
//      let h = document.getElementById("tqf-terms");
//      if (h && h.checked)
//        HiLighter.doHilight("msgLoaded", null, HiLighter.highlightBrowserElement);
//     }
//  },

  FolderDisplayListener: {
    onDisplayingFolder: function(aFolderDisplay) {
      let b = document.getElementById("qfb-qs-body");
      if (!b || !aFolderDisplay.active || !aFolderDisplay.displayedFolder)
        return

TotalQuickFilter.log.debug("onDisplayingFolder:  "+
                             aFolderDisplay.displayedFolder.prettyName);
      let disabled = aFolderDisplay.displayedFolder.server.filterScope ==
                       Ci.nsMsgSearchScope.newsFilter && !Services.io.offline;
      TotalQuickFilter.setTextFilterButton(b, null, disabled, null);
    },
    onLeavingFolder: function(aFolderDisplay) {
      let f = TotalQuickFilter.getTabFilterValues(aFolderDisplay);
      if (!aFolderDisplay.active || !f || f.sticky)
        return

TotalQuickFilter.log.debug("onLeavingFolder:  "+
                             aFolderDisplay.displayedFolder.prettyName);

      QuickFilterManager.clearAllFilterValues(f);
    }
  },

  TabMonitor: {
    monitorName: "TotalQuickFilter",
    onTabTitleChanged: function() {},
    onTabOpened: function(aTab, aIsFirstTab, aWasFirstTab) {
TotalQuickFilter.log.debug("onTabOpened: ");
    },
    onTabRestored: function(aTab, aState, aIsFirstTab) {
TotalQuickFilter.log.debug("onTabRestored: ");
    },
    onTabSwitched: function(aTab, aOldTab) {
TotalQuickFilter.log.debug("onTabSwitched: ");
      TotalQuickFilter.clearFilter();
      if (aTab._ext.quickFilter) {
TotalQuickFilter.log.debug("onTabSwitched: title - "+aTab.title);
         let t = document.getElementById("qfb-qs-textbox");
         TotalQuickFilter.setTextInput(t);
         TotalQuickFilter.setTextFilterBarButton(t, "qfb-text-filter");
       }
    }
  },

  AddonListener: {
    reset: false,
    resetSession: function(aAddon, aReset) {
      if (aAddon.id != TotalQuickFilter._id)
        return;
TotalQuickFilter.log.debug("AddonListener.resetSession: reset - "+aReset);
      this.reset = aReset;
    },
    onUninstalling: function(aAddon) { this.resetSession(aAddon, true); },
    onDisabling: function(aAddon) { this.resetSession(aAddon, true); },
    onEnabling: function(aAddon) { this.resetSession(aAddon, false); },
    onOperationCancelled: function(aAddon) {
      if (!(aAddon.pendingOperations & AddonManager.PENDING_DISABLE))
        this.resetSession(aAddon, false);
    }
  }

}; // TotalQuickFilter

var HiLighter = {
  HTML_NS: "http://www.w3.org/1999/xhtml",
  ButtonsArray: [],
  HilightRulesStrings: [],
  HilightRulesDefault: [
    {fg:"white",bg:"darkcyan",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"darkorange",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"mediumorchid",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"green",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"blue",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"red",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"goldenrod",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"indianred",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"midnightblue",b:false,u:false,i:false,f:"",s:""},
    {fg:"white",bg:"maroon",b:false,u:false,i:false,f:"",s:""}
  ],

  init: function() {
    // Set messagepane as a highlight target for load events, http only.
    let mp = document.getElementById("messagepane");
    mp.addEventListener("load", function onLoad(event) {
      if (["http:","https:"].indexOf(event.target.location.protocol) == -1)
        return;
TotalQuickFilter.log.debug("HiLighter.onLoad: http load - "+ event.target.location.href);
      HiLighter.doHilight("httpload", null, mp)
    }, true);

    // Set multimessage as a highlight target for load events.
//    mp = document.getElementById("multimessage");
//    mp.addEventListener("load", function onLoad(event) {
//TotalQuickFilter.log.debug("HiLighter.onLoad: multimessage load - "+ event.target.location.href);
//      if (!["chrome:"].includes(event.target.location.protocol))
//        return;
//      HiLighter.doHilight("multimessageload", null, mp)
//    }, true);

    this.registerStyles();
    this.registerHilightStyles();
  },

  // Allow overrides, to highlight any <browser> doc.
  highlightBrowserId: null,
  get highlightBrowserElement() {
    if (this.highlightBrowserId)
      return document.getElementById(this.highlightBrowserId);
    if (gFolderDisplay && gFolderDisplay.summarizeSelectionInFolder &&
        !gFolderDisplay.messageDisplay.singleMessageDisplay)
      return document.getElementById("multimessage");
    return document.getElementById("messagepane");
  },

  getContentWin: function(aBrowser) {
    let contentWin = aBrowser._finderHL._getWindow();
    if (!(contentWin instanceof Ci.nsIDOMWindow) || !contentWin.document.body ||
        contentWin.document.readyState != "complete")
      contentWin = null;
TotalQuickFilter.log.debug("HiLighter.getContentWin: win - "+contentWin);
    return contentWin;
  },

  // Get a Finder for the <browser>.
  getFinder: function(aBrowser) {
    if (aBrowser._finderHL)
      return aBrowser._finderHL;

TotalQuickFilter.log.debug("HiLighter.getFinder: new for - "+aBrowser.id);
    let Finder = Cu.import("resource://gre/modules/Finder.jsm", {}).Finder;
    let finder = new Finder(aBrowser.docShell);
    finder._Browser = aBrowser;
    finder._notifyHighlightFinished = function(aHighlight, aBrowser) {
      for (let l of this._listeners) {
        try {
          l.onHighlightFinished(aHighlight, this._Browser);
        } catch (ex) {}
      }
    }
    finder._highlightSleep = function(d) { return; }
    finder.addResultListener(HiLighter);

    return aBrowser._finderHL = finder;
  },

  get HilightRules() {
    if (this._rulesMap)
      return this._rulesMap;

    let prefName = TotalQuickFilter.prefBranch + ".hilightRules";
    let rulesMap;
TotalQuickFilter.log.debug("HiLighter.HilightRules: "+prefName);
    try {
      rulesMap = JSON.parse(Services.prefs.getCharPref(prefName));
    }
    catch (ex) {
      TotalQuickFilter.log.warn("HiLighter.HilightRules: pref reset or error, " +
                                "restoring defaults - " + prefName);
      // Either error or pref has been reset.  Store pref with default.
      Services.prefs.setCharPref(prefName, JSON.stringify(this.HilightRulesDefault));
      rulesMap = JSON.parse(Services.prefs.getCharPref(prefName));
    }

    return this._rulesMap = rulesMap;
  },

  set HilightRules(aRules) {
    this._rulesMap = null;
    let prefName = TotalQuickFilter.prefBranch + ".hilightRules";
TotalQuickFilter.log.debug("HiLighter.setHilightRules: "+prefName);

    try {
      if (aRules)
        Services.prefs.setCharPref(prefName, JSON.stringify(aRules));
      else
        Services.prefs.clearUserPref(prefName);
    }
    catch (ex) {
      TotalQuickFilter.log.warn("HiLighter.setHilightRules: error storing pref - " + prefName);
    }
  },

  _HilightMarkersMode: null,
  get HilightMarkersMode() {
    if (this._HilightMarkersMode !== null)
      return this._HilightMarkersMode;
    let b = document.getElementById("tqf-terms");
    return b && b.getAttribute("markers") == "true";
  },

  get HilightTermsCount() {
    return Object.keys(this.HilightRules).length;
  },

  // Total highlight words supported, for hilight toggle purposes (using css).
  // By default, there are 10 different styles, which may be repeated once, for
  // a total of 20.  This is so we can use css only.
  HilightTermsTotalCount: 20,

  getRuleIndex: function(aIndex) {
    return Math.abs(aIndex%this.HilightTermsCount);
  },

  getRuleString: function(aElement, aIndex) {
    if (this.HilightRulesStrings[aIndex]) {
      if (aElement.tagName == "toolbarbutton")
        return this.HilightRulesStrings[aIndex]
                   .replace(/(font-size:\s)(.*)(%;)/, "$1100$3");
      return this.HilightRulesStrings[aIndex];
    }

    let ruleStr = "";
    let ruleIndex = this.getRuleIndex(aIndex);
    // Use !important to override web page styles.
    for (let [name, value] in Iterator(this.HilightRules[ruleIndex])) {
      if (value)
        switch (name) {
          case "fg":
            ruleStr += "color: " + value + " !important; ";
            break;
          case "bg":
            ruleStr += "background-color: " + value + " !important; ";
            break;
          case "b":
            ruleStr += "font-weight: bold !important; ";
            break;
          case "u":
            ruleStr += "text-decoration: underline !important; ";
            break;
          case "i":
            ruleStr += "font-style: italic !important; ";
            break;
          case "f":
            ruleStr += "font-family: " + value + " !important; ";
            break;
          case "s":
            ruleStr += "font-size: " + value + "% !important; ";
            break;
          default:
            break;
        }
    }

    this.HilightRulesStrings[aIndex] = ruleStr;
    if (aElement.tagName == "toolbarbutton")
      return ruleStr.replace(/(font-size:\s)(.*)(%;)/, "$1100$3");
    return ruleStr;
  },

  get MarkersBackgroundPref() {
    return Services.prefs.getCharPref(TotalQuickFilter.prefBranch + ".markersBackground");
  },

  set MarkersBackgroundPref(val) {
    Services.prefs.setCharPref(TotalQuickFilter.prefBranch + ".markersBackground", val);
  },

  registerStyles: function() {
    let url = "chrome://totalquickfilter/skin/hl.css";
    let ssSvc = Cc["@mozilla.org/content/style-sheet-service;1"]
                  .getService(Ci.nsIStyleSheetService);
    let ssUri = Services.io.newURI(url, null, null);
    if (!ssSvc.sheetRegistered(ssUri, ssSvc.AGENT_SHEET))
      ssSvc.loadAndRegisterSheet(ssUri, ssSvc.AGENT_SHEET);
  },

  HilightRulesURI: null,
  registerHilightStyles: function() {
    let index = -1;
    let ruleMap, ruleStr, rule, rules = "", color;
    while (++index < this.HilightTermsTotalCount) {
      let ruleIndex = this.getRuleIndex(index);
      ruleStr = this.getRuleString({tagName:"style"}, ruleIndex);
      ruleMap = HiLighter.HilightRules[ruleIndex];
      color = ruleMap.bg == "transparent" ? ruleMap.fg : ruleMap.bg;
      rule = "body[hilight][tqf-term-" + index + "]  span.tqf-term-" + index +
             " { " + ruleStr + "} " +
             "div.hlMarker-tqf-term-style-" + index +
             " { " + "background-color: " + color + " !important; } ";
      rules += rule;
    }

TotalQuickFilter.log.debug("HiLighter.registerHilightStyles: cssRules - "+rules);
    let rulesUrl = "data:text/css;base64," + window.btoa(rules);
    let rulesUri = Services.io.newURI(rulesUrl, null, null);

    let ssSvc = Cc["@mozilla.org/content/style-sheet-service;1"]
                  .getService(Ci.nsIStyleSheetService);
    if (this.HilightRulesURI &&
        ssSvc.sheetRegistered(this.HilightRulesURI, ssSvc.AGENT_SHEET))
      ssSvc.unregisterSheet(this.HilightRulesURI, ssSvc.AGENT_SHEET);
    if (!ssSvc.sheetRegistered(rulesUri, ssSvc.AGENT_SHEET)) {
      ssSvc.loadAndRegisterSheet(rulesUri, ssSvc.AGENT_SHEET);
      this.HilightRulesURI = rulesUri;
TotalQuickFilter.log.debug("HiLighter.registerHilightStyles: registered ");
    }
  },

  HilightRulesZIndexes: {},
  HilightRulesZIndexLast: 0,
  HilightRulesZIndexURI: null,
  setHilightMarkersZStyles: function(aTermIndex) {
TotalQuickFilter.log.debug("HiLighter.setHilightMarkersZStyles: aTermIndex - "+aTermIndex);
    if (!this.HilightRulesZIndexes[aTermIndex])
      this.HilightRulesZIndexes[aTermIndex] = 0;
    let rules = "";
    for (let [index, ] in Iterator(this.HilightRulesZIndexes)) {
      if (index == aTermIndex)
        this.HilightRulesZIndexes[index] = ++this.HilightRulesZIndexLast;
      let rule = ".hlMarker-tqf-term-" + index +
                 " { " + "z-index: " + this.HilightRulesZIndexes[index] + "; } ";
      rules += rule;
    }

TotalQuickFilter.log.debug("HiLighter.setHilightMarkersZStyles: rules - "+rules);
    let rulesUrl = "data:text/css;base64," + window.btoa(rules);
    let rulesUri = Services.io.newURI(rulesUrl, null, null);

    let ssSvc = Cc["@mozilla.org/content/style-sheet-service;1"]
                  .getService(Ci.nsIStyleSheetService);
    if (this.HilightRulesZIndexURI &&
        ssSvc.sheetRegistered(this.HilightRulesZIndexURI, ssSvc.AGENT_SHEET))
      ssSvc.unregisterSheet(this.HilightRulesZIndexURI, ssSvc.AGENT_SHEET);
    if (!ssSvc.sheetRegistered(rulesUri, ssSvc.AGENT_SHEET)) {
      ssSvc.loadAndRegisterSheet(rulesUri, ssSvc.AGENT_SHEET);
      this.HilightRulesZIndexURI = rulesUri;
TotalQuickFilter.log.debug("HiLighter.setHilightMarkersZStyles: set ");
    }
  },

  updateHilightButtons: function(aWhat) {
    let buttonsBar = TotalQuickFilter.getButtonsBar();
    if (!buttonsBar)
      return;
    for (let button of buttonsBar.childNodes) {
      if (button.tagName == "toolbarbutton") {
        if (aWhat.style && !button.hasAttribute("negated")) {
          let ruleIndex = this.getRuleIndex(parseInt(button.getAttribute("index")));
          button.setAttribute("style", this.getRuleString(button, ruleIndex));
        }
        if (aWhat.counts)
          button.label = button.getAttribute("displayterm");
      }
    }
  },

  doHilight: function(aCommand, aButton, aBrowser) {
TotalQuickFilter.log.debug("HiLighter.doHilight: command - "+ aCommand);
    if (!aButton)
      // Reset counts if highlighting all buttons.
      HiLighter.updateHilightButtons({counts:true});

    if (!aBrowser || !TotalQuickFilter.isHilighterStateActive())
      return;

    let contentWin = aBrowser.contentWindow;
    if (!(contentWin instanceof Ci.nsIDOMWindow) || !contentWin.document.body ||
        contentWin.document.readyState != "complete")
      return;

    let hlButton = document.getElementById("tqf-terms");
    if (!hlButton || hlButton.hasAttribute("disabled"))
      return;

    if (aButton) {
      // Single button.
      if (aButton.hasAttribute("negated"))
        return;

      if (!aButton.checked) {
        HiLighter.hilightHide(aButton, aBrowser);
        return;
      }

      HiLighter.ButtonsArray = [aButton];
      HiLighter.setHilightMarkersZStyles(parseInt(aButton.getAttribute("index")));
    }
    else {
      // All buttons.
      let buttonsBar = TotalQuickFilter.getButtonsBar();
      if (!buttonsBar || !buttonsBar.childElementCount || buttonsBar.collapsed)
        return;

      let buttonsArray = [];
      for (let button of buttonsBar.childNodes) {
        if (aCommand != "click")
          button.removeAttribute("hilit");
        if (button.tagName == "toolbarbutton" && button.checked &&
            !button.hasAttribute("negated")) {
          buttonsArray.push(button);
        }
      }

      if (!buttonsArray.length)
        return;

      HiLighter.ButtonsArray = buttonsArray;
    }

    hlButton.setAttribute("busy", true);
    contentWin.document.body.setAttribute("hilight", true);
    HiLighter.hilighter(aBrowser);
  },

  // Hilight search terms.
  hilighter: function(aBrowser) {
    let button = this.ButtonsArray[0];
    if (!button) {
TotalQuickFilter.log.debug("HiLighter.hilighter: no more buttons ");
      document.getElementById("tqf-terms").removeAttribute("busy");
      return;
    }

    let term = button.getAttribute("term");
TotalQuickFilter.log.debug("HiLighter.hilighter: term - "+ term);
    // Use selection hilight to get the elements and wrap them in a <span>
    // element for multiselection hilighting.  Hilighting full or partial
    // text broken up by or within other html tags (<a>, <b>, etc.) is
    // possible with this method.
    let finder = this.getFinder(aBrowser);
    finder.caseSensitive = !TotalQuickFilter.IgnoreCase;
//    finder.highlight(true, term);
    setTimeout(function() { finder.highlight(true, term); }, 100);
  },

  // Hilight completed listener.
  onHighlightFinished: function(aHighlight, aBrowser) {
TotalQuickFilter.log.debug("HiLighter.onHighlightFinished: ");

    let contentWin = HiLighter.getContentWin(aBrowser);
    let button = HiLighter.ButtonsArray.shift();
    let term = button.getAttribute("term");
    let controller = HiLighter.getFinder(aBrowser)._getSelectionController(contentWin);
    if (!controller) {
TotalQuickFilter.log.info("HiLighter.onHilightFinished: could not get controller");
      return;
    }
    let selection = controller.getSelection(Ci.nsISelectionController.SELECTION_FIND);
    let rangeCount = selection.rangeCount;
TotalQuickFilter.log.debug("HiLighter.onHighlightFinished: process term:rangeCount - "+
                           term+":"+rangeCount);

    let counts = "";
    let termIndex = parseInt(button.getAttribute("index"));
    let termId = "tqf-term-" + termIndex;
    let termStyleId = "tqf-term-style-" + HiLighter.getRuleIndex(termIndex);
    if (gFolderDisplay.selectedCount)
      counts = " (" + (rangeCount ? rangeCount : "?") + ")";
    button.label = button.getAttribute("displayterm") + counts;

    if (button.hasAttribute("hilit")) {
      selection.removeAllRanges();
    }
    else {
      // Only if not already done.
      let hlnode = document.createElementNS(HiLighter.HTML_NS, "span");
      hlnode.classList.add("tqf-terms");
      hlnode.classList.add(termId);
      hlnode.classList.add(termStyleId);
      hlnode.setAttribute("index", termIndex);
      while (rangeCount) {
        // Go backwards to process ranges in a node, to preserve offsets..
        let range = selection.getRangeAt(--rangeCount);
        if (!range.startContainer.parentNode.classList.contains(termId)) {
          let node = hlnode.cloneNode(false);
          // surroundContents(node) is/was unreliable, esp cross link spans.
//          range.surroundContents(hlnode);
          node.appendChild(range.extractContents());
          range.insertNode(node);
        }
        selection.removeRange(range);
      }
      button.setAttribute("hilit", true);
    }

    contentWin.document.body.setAttribute(termId, true);

    if (HiLighter.HilightMarkersMode)
      HiLighter.hilightMarker(button, contentWin, aBrowser);

    // Next button..
    HiLighter.hilighter(aBrowser);
  },

  /**
   * The hilight marker code is inspired by the Search Marker extension by
   * Matt Kenison<matt@penguinus.com> http://www.penguinus.com/dev/searchmarker/
   */

  // Hilight button rt click.
  doHilightMarkersAll: function(aBrowser) {
    let contentWin = HiLighter.getContentWin(aBrowser);
    if (!contentWin) // || gFolderDisplay.selectedCount != 1)
      return;

    let buttonsBar = TotalQuickFilter.getButtonsBar();
    if (!buttonsBar || !buttonsBar.childElementCount || buttonsBar.collapsed)
      return;

    for (let button of buttonsBar.childNodes) {
      if (button.tagName == "toolbarbutton" && button.checked &&
          !button.hasAttribute("negated"))
        HiLighter.hilightMarker(button, contentWin, aBrowser);
    }
  },

  hilightMarker: function(aTermButton, aContentWin, aBrowser) {
TotalQuickFilter.log.debug("HiLighter.hilightMarker: START buttonId - "+ aTermButton.id);

    // Calculate space taken by scrollbar buttons, markers should not be above
    // or below the scroll track (if scrollbar buttons are visible).  Can't get
    // real height of button, so get scrollbar width and assume they're square.
    // XXX:  This doesn't work for scrollbar buttons moved together style
    // (OSX e.g.) or buttons not visible.
TotalQuickFilter.log.debug("HiLighter.hilightMarker: CONTINUE widths - "+
                           aBrowser.clientWidth+":"+aBrowser.scrollWidth+":"+
                           aBrowser.boxObject.width+":"+
                           aContentWin.document.documentElement.clientWidth);
    let width = aBrowser.clientWidth - aContentWin.document.documentElement.clientWidth;
    if (!width)
      // No need for markers unless there's a scrollbar.
      return;

    let body = aContentWin.document.body;
    let markerBox = aContentWin.document.getElementById("hlMarkerBox");
    let markerBar = aContentWin.document.getElementById("hlMarkerBar");
    let termIndex = parseInt(aTermButton.getAttribute("index"));
    let termId = "tqf-term-" + termIndex;
    let markerId = "hlMarker-" + termId;
    let markerStyleId = "hlMarker-tqf-term-style-" + this.getRuleIndex(termIndex);
    let searchResults = []

    if (markerBox) {
      markerBox.removeAttribute("hidden");
      markerBar.setAttribute(markerId, true);
    }

    // Find all hilighted, not previously marked <span> entries.
    for (let span of body.getElementsByClassName(termId)) {
      if (!span.hasAttribute("marked"))
        searchResults.push(span);
    }
TotalQuickFilter.log.trace("HiLighter.hilightMarker: scrollbar width:count - "+
                           width+":"+searchResults.length);
    if (!searchResults.length)
      // No need for markers unless there are hilights.
      return;

    // Create element to contain the markers.
    if (!markerBox) {
      markerBox = document.createElementNS(HiLighter.HTML_NS, "div");
      markerBox.setAttribute("id", "hlMarkerBox");
      markerBox.setAttribute("style", "background-color: " +
                                      this.MarkersBackgroundPref + ";");
      markerBar = document.createElementNS(HiLighter.HTML_NS, "div");
      markerBar.setAttribute("id", "hlMarkerBar");
      markerBar.setAttribute("style", "top: " + width + "px; " +
                                      "height: calc(100% - " + width*2 + "px);");
      markerBox.appendChild(markerBar);
      body.appendChild(markerBox);
    }
TotalQuickFilter.log.trace("HiLighter.hilightMarker: termId:count - "+
                           termId+":"+searchResults.length);

    let absoluteTop, markerPos, markerNode;
    let markerDiv = document.createElementNS(HiLighter.HTML_NS, "div");
    markerDiv.classList.add("hlMarkers");
    markerDiv.classList.add(markerId);
    markerDiv.classList.add(markerStyleId);

    function marker(searchResult) {
      if (!searchResult || !HiLighter.HilightMarkersMode || !aTermButton.checked)
        return;

      aContentWin.setTimeout(function() {
        // Get the abs y location of term as a percentage of the visible area.
        absoluteTop = HiLighter.getAbsoluteTop(searchResult);
        markerPos = absoluteTop / body.parentNode.scrollHeight * 100;
        if (Number.isNaN(markerPos))
          // Just catch /0 here, in content transition cases.
          return;

        markerPos = Math.round(markerPos * 100) / 100;

        // Create the marker element.
        markerNode = markerDiv.cloneNode(false);
        markerNode.setAttribute("title", searchResult.textContent);
        markerNode.setAttribute("style", "top: " + markerPos + "% !important;");
        markerNode.setAttribute("absolutetop", absoluteTop);
        markerNode.addEventListener("click", HiLighter.scrollTo);
        markerBar.appendChild(markerNode);
        markerBar.setAttribute(markerId, true);

        searchResult.setAttribute("marked", true);
        marker(searchResults.shift());

      }, 0); // setTimeout

    } // marker

    marker(searchResults.shift());
  },

  // Recursively add the relative top of the parent element.
  getAbsoluteTop: function(element) {
    if (element == null)
      return 0;
    return this.getAbsoluteTop(element.offsetParent) + element.offsetTop;
  },

  // Go to the marker's page position.
  scrollTo: function(aEvent) {
    let y = parseInt(aEvent.target.getAttribute("absolutetop"));
TotalQuickFilter.log.trace("HiLighter.scrollTo: y - "+y);
    aEvent.target.ownerDocument.defaultView.scrollTo(0, y);
  },

  hilightHide: function(aButton, aBrowser) {
    let contentWin = HiLighter.getContentWin(aBrowser);
    if (!contentWin)
      return;

    let body = contentWin.document.body;
    let markerBox = contentWin.document.getElementById("hlMarkerBox");
    let markerBar = contentWin.document.getElementById("hlMarkerBar");
    if (aButton) {
      let termIndex = parseInt(aButton.getAttribute("index"));
      let termId = "tqf-term-" + termIndex;
      let markerId = "hlMarker-" + termId;
TotalQuickFilter.log.debug("HiLighter.hilightHide: termId - "+termId);
      aButton.label = aButton.getAttribute("displayterm");
      body.removeAttribute(termId);
      if (markerBox)
        markerBar.removeAttribute(markerId);
    }
    else {
TotalQuickFilter.log.debug("HiLighter.hilightHide: all ");
      body.removeAttribute("hilight");
      if (markerBox)
        markerBox.setAttribute("hidden", true);
    }
  },

  hilightMarkersClear: function(aBrowser) {
    let contentWin = HiLighter.getContentWin(aBrowser);
    if (!contentWin)
      return;

    let markerBox = contentWin.document.getElementById("hlMarkerBox");
    if (markerBox)
      markerBox.setAttribute("hidden", true);
  }

}; // HiLighter
